package planet.behaviour;

import java.util.Vector;
import java.util.Iterator;

import planet.commonapi.exception.InitializationException;
import planet.commonapi.RouteMessage;
import planet.commonapi.Node;
import planet.util.Properties;
import planet.behaviour.exception.NoBehaviourDispatchedException;
import planet.behaviour.exception.NoSuchBehaviourException;
import planet.behaviour.exception.OutOfRangeError;

/**
 * The BehavioursPool's class is aimed at providing an internal  scheduler of
 * node's behaviuours. At startup, behaviours are registered to the pool from 
 * Behaviour.properties file, following the next syntax:
 * <br><pre>    
 *      Class = Type, Mode, Probability, When, Role 
 * </pre>
 * Then, a message interceptor is build up and is ready to invoke  behaviours
 * when a message pattern matches to those provided by <b> Behaviour.properties </b>
 * file.  
 * @author <a href="mailto: marc.sanchez@estudiants.urv.es">Marc Sanchez</a>
 * Date: 10/10/2004   
 */
public class BehavioursPoolImpl implements BehavioursPool {
	/**
	 * BehavioursPool properties: mapping from <patterns> to <behaviours>.
	 * 		~ Patterns are composed by: Type and Mode, for example: 
	 * 					"QUERY_JOIN" and "REQUEST".
	 * 		~ Types: NodeImpl class has several public static final int constants
	 * 				 from range 0..NumberOfTypes.
	 * 		~ Modes: NodeImpl class has several public static final int constants
	 * 				 from range 0..NumberOfModes.
	 * 	By reflection, the behaviour's pool given a set of patterns P from
	 *  <b>behaviours.properties</b> file do:
	 *  
	 *  	For each pattern on P do
	 *  		Mode = pattern.getMode();
	 *  		Type = pattern.getType();
	 *  		beh[Type * NumberOfModes + Mode] = pattern; 
	 *      done;
	 *  
	 *  Moreover, Pattern's definiton provides additional specifiers:
	 *  	
	 *      ~ Incoming RouteMessage's destination: Local, Remote  and
	 *  	Always. Local qualifier affects iff incoming messages for
	 *      the local node. Hence, only behaviours for the local node
	 *      may be matched. Remote qualifier affects iff messages for
	 *      remote nodes. Hence, programmer must code behaviours which 
	 *      alter the underlying overlay protocol when RouteMessages
	 *      are destined to remote nodes. Always modifier is used as
	 *      a wild card to ignore local or remote behaviour.   
	 *      
	 *      ~ Behaviours's role: Bad, Good and Neutral. Bad qualifier
	 *      alters only the behaviours from bad nodes. Good qualifier
	 *      alters only the behaviours from good nodes and Neutral is
	 *      a wild card used to ignore peer's behaviour. 
	 *   
	 **/
	protected java.util.Vector[][] beh;
	/**
	 * Behaviour's Property: Good modifier for Behaviour's role. This means the behaviour
	 * qualified by this modifier only will be scheduled by good peers.
	 */
	protected static int GoodRole       = 0;
	/**
	 * Behaviour's Property: Bad modifier for Behaviour's role. This means the behaviour
	 * qualified by this modifier only will be scheduled by bad peers.
	 */
	protected static int BadRole        = 1;
	/**
	 * Behaviour's Property: Local modifier applied to a behaviour means it only will
	 * be scheduled when a RouteMessage has as destination the local node.
	 */
	protected static int LocalTraffic   = 0;
	/**
	 * Behaviour's Property: Remote modifier applied to a behaviour means it only will
	 * be scheduled when a RouteMessage has as destination a remode node and needs to
	 * be rerouted trough the strucutred overlay again.
	 */
	protected static int RemoteTraffic  = 1;
	/*
	 * Class descriptor for node's class.
	 */
	protected Class nodeClass;
	/*
	 * Number of modes for node's RouteMessageImpl.
	 */
	protected int numberModes;
	/*
	 * Number of types for node's RouteMessageImpl.
	 */
	protected int numberTypes;
	/*
	 * Filter instance.
	 */
	protected planet.behaviour.Filter filter;
	/*
	 * Number of behaviour's slots: numberTypes * numberModes;
	 */
	protected int behSlots;
	/*
	 * Behaviour's Property: Always modifier applied to a behaviour ignores the
	 * distinction between local or remote traffic handled by a node.
	 */
	protected static final int ALWAYS_MASK = 0x2;
	/*
	 * Behaviour's Property: Neutral modifier applied to a behaviour ignores the
	 * distinction between local or remote traffic handled by a node.
	 */
	protected static final int NEUTRAL_MASK = 0x1;
	/*
	 * Behaviour's Property: Lack Of Neutral and Always Mask. 
	 */
	protected static final int NULL_MASK = 0x0;
	/*
	 * Total number of Behaviour's Queue: 
	 * 		~ Good and Local  Behaviours: 0
	 * 		~ Good and Remote Behaviours: 1
	 * 		~ Bad  and Local  Behaviours: 2
	 * 		~ Bad  and Remote Behaviours: 3
	 */
	protected static final int BEH_MAPPINGS = 4;
	
	
	private static final String Queues[] = { 
			"Good and Local  ", 
			"Good and Remote ", 
			"Bad  and Local  ", 
			"Bad  and Remote " }; 
	
	public BehavioursPoolImpl(Class filterClass, java.util.Vector patterns, Class nodeClass) throws InitializationException {
		// Filter
		if (filterClass != null)
			try {
				 filter = (planet.behaviour.Filter) filterClass.newInstance();
			} catch(Exception e) {
				throw new InitializationException("Filter Instance Failed For '" + filterClass.getName() + "'.", e);
			}
		// Behaviour's list
		this.nodeClass = nodeClass;
		
		java.lang.reflect.Field NumberOfModes = null;
		java.lang.reflect.Field NumberOfTypes = null;
		try {
			 NumberOfModes = nodeClass.getField(BehavioursPool.NumberOfModes);
			 NumberOfTypes = nodeClass.getField(BehavioursPool.NumberOfTypes);
		} catch(NoSuchFieldException e) {
			throw new InitializationException("Node Class provided does not has mandatory properties: \"NumberOfModes\" and \"NumberOfTypes\"", e);
		}
		try {
			numberModes = NumberOfModes.getInt(null);
			numberTypes = NumberOfTypes.getInt(null);
		} catch(Exception e) {
			throw new InitializationException("Node Class provided does not provide public static int properties: \"NumberOfModes\" and \"NumberOfTypes\"", e);
		}
		if (numberModes <= 0) throw new InitializationException("Node Class provided does not provide enough RouteMessage's modes");
		if (numberTypes <= 0) throw new InitializationException("Node Class provided does not provide enough RouteMessage's types");
		
		behSlots = numberTypes * numberModes;
		// Mapping From <Pattern> To <Behaviour>
		this.build(patterns);
		
		if (((BehaviourProperties) Properties.behProps).debugInfo) prettyPrintAll();
	}
	/**
	 * The WhichQueue's method selects for a given pattern what Queues must
	 * include a copy of the current behaviour.
	 * @param Mask The Mask for Pattern.Always and Pattern.Neutral modifiers.
	 * @param RoleOf The Behaviour's Role for a <b>Good</b> or <b>Bad</b> peer.
	 * @param WhenTo Specifies Behaviour should be activated when incoming messages 
	 * refer to the local node or either when refer to a remote.
	 * @return Which Behaviour's Queue must copy this behaviour.
	 */
	protected boolean[] whichQueues(int Mask, int RoleOf, int WhenTo) {
		boolean[] which = { false, false, false, false };
		if (Mask == NULL_MASK) which[RoleOf * BEH_MAPPINGS / 2 + WhenTo] = true;
		else if ((Mask & (NEUTRAL_MASK|ALWAYS_MASK)) == (NEUTRAL_MASK|ALWAYS_MASK))
			which = new boolean[] { true, true, true, true };
		else if ((Mask & NEUTRAL_MASK) == NEUTRAL_MASK)  {
			which[GoodRole * BEH_MAPPINGS / 2 + WhenTo] = true;
			which[BadRole  * BEH_MAPPINGS / 2 + WhenTo] = true;
		} else if ((Mask & ALWAYS_MASK) == ALWAYS_MASK) {
			which[RoleOf * BEH_MAPPINGS / 2 + LocalTraffic]  = true;
			which[RoleOf * BEH_MAPPINGS / 2 + RemoteTraffic] = true;
		}
		return which;
	}
	/**
	 * Given the patterns sorted from more-to-less specific as input,
	 * builds a mapping from RouteMessage's patterns to behaviours.
	 * @param patterns The List of patterns.
	 */
	protected void build(Vector patterns) throws InitializationException {
		beh = new Vector[BEH_MAPPINGS][behSlots];
		for (int behQueue = 0; behQueue  < BEH_MAPPINGS; behQueue++)
			for(int slot = 0; slot < behSlots; slot++)
				beh[behQueue][slot] = new Vector();
		
		// Flags	
		int Flags;
		// Role Modifier
		int role = 0;
		// When Modifier
		int when = 0;
		
		while(!patterns.isEmpty()) {
			Pattern pattern = (Pattern) patterns.remove(0);
			Flags = NULL_MASK;

			// Role
			String  roleOf  = pattern.getRoleOf();
			if (roleOf.equals(Pattern.Good)) { 
				role = GoodRole;
			} else if(roleOf.equals(Pattern.Bad)) {
				role = BadRole;
			} else Flags = Flags | NEUTRAL_MASK;

			// When
			String  whenTo  = pattern.getWhenTo();
			if (whenTo.equals(Pattern.Local)) {
				when = LocalTraffic;
			} else if(whenTo.equals(Pattern.Remote)) {
				when = RemoteTraffic;
			} else Flags = Flags | ALWAYS_MASK;
			
			String  typeOf = pattern.getTypeOf();
			String  modeOf = pattern.getModeOf();
			boolean[] queues = whichQueues(Flags, role, when);

			
			for(int queue = 0; queue < BEH_MAPPINGS; queue++) {
				if (!queues[queue]) continue;
				// Type's is '?' wildcard
				if(typeOf.equals(Pattern.Cwildcard)) {
					for (int typePos = 0; typePos < behSlots; typePos = typePos + numberModes) {
						boolean empty = true;
						for (int modePos = 0; modePos < numberModes; modePos++)
							if (beh[queue][modePos + typePos].size() > 0) { 
								empty = false; 
								break;
							}
						if (empty) setPattern(queue, typePos, modeOf, pattern);
					}
					// 	Type's is '*' wildcard
				} else if(typeOf.equals(Pattern.Uwildcard)) {
					for (int pos = 0; pos < behSlots; pos = pos + numberModes)
						setPattern(queue, pos, modeOf, pattern);
					// 	Type's tag
				} else {
					int typePos = 0;
					try {
						typePos = ((java.lang.reflect.Field) nodeClass.getField(typeOf)).getInt(null);
					} catch(Exception e) {
						throw new InitializationException("Node class [" + nodeClass.getName() + "] does " +
							"not provide constant [" + typeOf + "]", e);
					}
					if (inRangeOf(numberTypes, typePos)) setPattern(queue, typePos * numberModes, modeOf, pattern);
					else throw new InitializationException("Type [" + typeOf + "] constant not in range [" + 0 + ".." + numberTypes + "]");
				}
			}
		}
	}
	/**
	 * Given the initial position for type on mapping <b>typePos</b> and <b>mode</b> as input, sets the 
	 * pattern on the mapping taking into account the grammar tree extracted from wildcards semantics.
	 * Hence, the maximum number of mappings goes from <b>typePos</b> to <b>typePos + NumberOfModes</b>.
	 * For example, suppose type is QUERY_JOIN and modes are REQUEST and REPLY the maximun number of
	 * mappings to set is 2;
	 * @param queue Identifies where to be queued the pattern.
	 * @param typePos Initial position on behaviours mapping array, i.e, [typePos..typePos + NumberOfModes]
	 * @param modeOf Mode of the pattern: Universal wildcard or '*', Complementary wildcard  or '?' and Tag.
	 * @param pattern Pattern to be treat.
	 */
	protected void setPattern (int queue, int typePos, String modeOf, Pattern pattern) throws InitializationException {
		// Type's is '?' wildcard
		if(modeOf.equals(Pattern.Cwildcard)) {
			for(int pos = typePos; pos < (typePos + numberModes); pos++) {
				if (beh[queue][pos].isEmpty()) beh[queue][pos].add(newInvoker(pattern));
			}
		// Type's is '*' wildcard
		} else if(modeOf.equals(Pattern.Uwildcard)) {
			for(int pos = typePos; pos < (typePos + numberModes); pos++) 
					beh[queue][pos].add(newInvoker(pattern));
		// Type's tag
		} else {
			int modePos = 0;
			try {
				modePos = ((java.lang.reflect.Field) nodeClass.getField(modeOf)).getInt(null);
			} catch(Exception e) {
				throw new InitializationException("Node class [" + nodeClass.getName() + "] does " +
						"not provide this constant: '" + modeOf + "'", e);
			}
			if (inRangeOf(numberModes, modePos)) beh[queue][typePos + modePos].add(newInvoker(pattern));
			else throw new InitializationException("Mode [" + modeOf + "] constant not in range [" + 0 + ".." + numberModes + "]");
		}		
	}
	
	private BehaviourInvoker newInvoker(Pattern pattern) throws InitializationException {
		Class behClass = pattern.getClassOf();
		try {
			return new BehaviourInvoker((Behaviour) behClass.newInstance(), pattern.getPdf());
		} catch(Exception e) {
			throw new InitializationException("Initialization Error For Behaviour '" + behClass.getName() + "'.", e);
		}
	}
	
	protected void prettyPrintAll() {
		for (int queue = 0; queue < BEH_MAPPINGS; queue++) {
			System.out.println("\n____ " + Queues[queue] + "_______________________________________________\n");
			printMap(queue);
			System.out.println("\n____________________________________________________________________");
		}
	}

	protected void printMap(int queue) {
		for (int slot = 0; slot < behSlots; slot++) {
			java.util.Iterator it = beh[queue][slot].iterator();
			if (it.hasNext()) System.out.print("\n[" + slot + "]: ");
			while(it.hasNext()) {
				BehaviourInvoker behInvoker = (BehaviourInvoker) it.next();
				String behaviour = behInvoker.getName(); 
				System.out.print(behaviour.substring(behaviour.lastIndexOf('.') + 1));
				if (it.hasNext()) System.out.print(", ");
				else System.out.print("\n");
			}
		}
	}
	/**
	 * Given an integer expression and upper bound as input, inRangeOf checks wether 
	 * this integer literal belongs to [0..upperBound).
	 * @param upperBound The upper bound of the set range.
	 * @param toEval Integer literal to evaluate.
	 * @return True if integer <b>toEval</b> belongs to [0..upperBound].
	 */
	private boolean inRangeOf(int upperBound, int toEval) {
		return toEval >= 0 && toEval < upperBound;
	}
	/**
	 * Given a RouteMessage and a Node as input, onMessage's method intends to
	 * invoke some behaviours only if RouteMessage's type and mode fields matches
	 * some behaviour's pattern.  
	 * @param msg RoteMessage taken as input.
	 * @param node Node taken as input.
	 * @return Returns either an array of RouteMessages or null when no messages
	 * need to transmit this node.
	 * @throws NoSuchBehaviourException when no behaviour matches RouteMessage's
	 * pattern. 
	 */
	public RouteMessage[] onMessage(RouteMessage msg, Node node) 
										throws NoSuchBehaviourException, NoBehaviourDispatchedException {
			// ______________Checking RouteMessage's Type and Mode Pattern______________________________
			if (!inRangeOf(numberModes, msg.getMode())) throw new OutOfRangeError("Mode [" + msg.getMode() + "] of RouteMessage Out of Range [0.." + numberModes + ") \n" + msg);
			if (!inRangeOf(numberTypes, msg.getType())) throw new OutOfRangeError("Type [" + msg.getType() + "] of RouteMessage Out of Range [0.." + numberTypes + ") \n" + msg);
			
			// ______________Filtering__________________________________________________________________
			boolean toReject = false;
			if (filter != null) { 
				toReject = filter.filter(msg, node);
			}
			if (toReject) return null;
			
			// _____________Dispatch Behaviour__________________________________________________________
			int whichQueue = node.playsGoodRole()? GoodRole: BadRole; 
			whichQueue = whichQueue * BEH_MAPPINGS / 2 + (node.isLocalMessage(msg)? LocalTraffic: RemoteTraffic);
			
			Vector behaviours = beh[whichQueue][msg.getType() * numberModes + msg.getMode()];

			if (behaviours.isEmpty()) throw new NoSuchBehaviourException();
			
			boolean anyExec = false;
			Vector replySet = new Vector(); 
			Iterator it = behaviours.iterator();
			while(it.hasNext()) {
				BehaviourInvoker behInvoker = (BehaviourInvoker) it.next();
				if (((BehaviourProperties) Properties.behProps).debugInfo)
					System.out.println("Invoking [" + behInvoker.getName() + "] On Node [" + node.getId() + "] " +
						"For [" + planet.simulate.Globals.typeToString(msg.getType()) + "]" +
							"[" + planet.simulate.Globals.modeToString(msg.getMode()) + "]");
				BehaviourInvokerStatus status = (BehaviourInvokerStatus) behInvoker.invoke(msg, node);
				anyExec |= status.isInvoked();
				if (status.getResultRouteMessages() == null) continue;
				copyInto(replySet, status.getResultRouteMessages());
			}
			if (!anyExec) throw new NoBehaviourDispatchedException();
			if (replySet.isEmpty()) return null;
			else return copyInto(replySet);				
	}
	
	protected void copyInto(Vector copy, RouteMessage[] source) {
		for(int i = 0; i < source.length; i++)
			copy.add(source[i]);
	}
	
	protected RouteMessage[] copyInto(Vector s) {
		RouteMessage[] copy = new RouteMessage[s.size()];
		Iterator it = s.iterator();
		for(int i = 0; it.hasNext(); i++)
			copy[i] = (RouteMessage) it.next();
		return copy;
	}
}
