package planet.commonapi;
import java.io.*;

import java.lang.Comparable;
//import java.util.*;

/**
 * This interface is an abstraction of an Id (or key) from the CommonAPI paper.
 * <br><br>
 * Any class that extends this <b>Id</b> for a specific overlay, must to offer
 * all theese constructors,  based with its argument list:
 * <ol>
 * <li><b>empty</b>: This constructor must to build always an Id with
 * the initial value, e.a. a zero value in most of cases.</li>
 * <li><b>int</b>: This constructor uses the int value to build a new Id.
 * For a specific overlay Id implementation.</li>
 * <li><b>double</b>: This constructor uses the double value to build a new Id.
 * For a specific overlay Id implementation.</li>
 * <li><b>byte[]</b>: This constructor uses the byte[] to build the internal
 * value of the new Id. This argument must to be according
 * the related Id.toByteArray() return value. For all overlay Id implementation.</li>
 * <li><b>int[]</b>: This constructor uses the int[] to build the internal
 * value of the new Id. For a specific overlay Id implementation.</li>
 * <li><b>String</b>: This constructor obtains from the String the 
 * internal representation of the new Id. For all overlay Id implementation.</li>
 * <li><b>BigInteger</b>: This constructor obtains from the BigInteger
 * the internal value of the new Id. For a specific overlay Id implementation.</li>
 * <li><b>Random</b>: This constructor uses the java.util.Random instance to build
 * a random Id.</li>
 * <li><b>String material, String algorithm</b>: The result obtained from the Hash <b>algorithm</b>
 * to the material, is used as the internal value for the new Id. Possible values of 
 * <b>algorithm</b> are "SHA-1" or "MD5", for example.</li>
 * </ol>
 * If the above constructors someone is nonapplicable to any Id implementation,
 * an InitializationException must to be thrown.
 *
 * @author <a href="mailto: cpairot@etse.urv.es">Carles Pairot</a>
 * @author <a href="mailto: ruben.mondejar@estudiants.urv.es">Ruben Mondejar</a>
 * @author <a href="mailto: jordi.pujol@estudiants.urv.es">Jordi Pujol</a>
 * 
 * @see planet.commonapi.exception.InitializationException
 */
public abstract class Id implements Comparable, Serializable {

  /**
   * Checks if this Id is between two given ids ccw (inclusive) and cw (exclusive) on the circle
   *
   * @param ccw the counterclockwise id
   * @param cw the clockwise id
   * @return true if this is between ccw (inclusive) and cw (exclusive), false otherwise
   */
  public abstract boolean between (Id ccw, Id cw);
  
  /**
   * Checks if this Id is in interval [init,end)
   *
   * @param init Id that open the interval
   * @return end Id that close the interval
   */
  public boolean Ebetween (Id init, Id end) {
  	if (this.equals (init)) {
        return true;
    }
    else {
      return this.between (init, end);
    }
  }
  
  /**
   * Checks if this Id is in interval (init,end]
   *
   * @param init Id that open the interval
   * @return end Id that close the interval
   */
  public boolean betweenE (Id init, Id end) {
  	if (this.equals (end)) {
  		return true;
  	}
  	else {
  		return this.between (init, end);
  	}
  }

  /**
   * Checks to see if the Id nid is clockwise or counterclockwise from this, on the ring. An Id is
   * clockwise if it is within the half circle clockwise from this on the ring. An Id is considered
   * counter-clockwise from itself.
   *
   * @param nid The id to compare to
   * @return true if clockwise, false otherwise.
   */
  public abstract boolean clockwise(Id nid);

  /**
   * Returns an Id corresponding to this Id plus a given distance
   *
   * @param offset the distance to add
   * @return the new Id
   */
  public abstract Id add (Id offset);

  /**
   * Returns an Id corresponding to this Id minus a given distance
   *
   * @param offset the distance to subtract
   * @return the new Id
   */
  public abstract Id subtract (Id offset);

  /** Shift operator. shift(-1,0) multiplies value of this by two, shift(1,0) divides by 2
   *
   * @param cnt the number of bits to shift, negative shifts left, positive shifts right
   * @param fill value of bit shifted in (0 if fill == 0, 1 otherwise)
   * @return this
   */
  public abstract Id shift (int cnt, int fill);

  /**
   * Returns a (mutable) byte array representing this Id
   *
   * @return A byte[] representing this Id
   */
  public abstract byte[] toByteArray();

  /**
   * Returns a string representing the full length of this Id.
   *
   * @return A string with all of this Id
   */
  public abstract String toStringFull();
  
  /**
   * Sets the new value with Id <b>newValue</b>.
   * <br><br>
   * This invokes:<br>
   * <center><b>setValue(newValue.getValue());</b></center>
   * @see planet.commonapi.Id#setValue(planet.commonapi.Id)
   * @param newValue Id with the new value.
   */
  public void setValue(Id newValue) {
  	setValue(newValue.getValue());
  }
  
  /**
   * Sets the new value to this Id with <b>newValue</b>
   * @param newValue Internal representation with the new value 
   * to set to this Id.
   */
  public abstract void setValue(Object newValue);
  
  /**
   * Returns the internal representation value of this Id.
   * @return The internal representatio value of this Id.
   */
  public abstract Object getValue();
  
  /**
   * This method returns the maximum value for an Id that is 
   * possible to build. Must to be reimplemented for any 
   * specific overlay Id. We recommend to use the singleton design pattern.
   * @return The maximum value sets to an Id. In this case,
   * always null. 
   */ 
  public static Id getMaximum() 
  {return null; }
  
  /**
   * Divides the maximum domain of the node Id in <b>numberOfNodes</b>,
   * to offers the offset between two consecutive nodes.
   * <br><br>
   * If the domain is [MIN..MAX], the aritmethical operation will be as follows:
   * <br>
   * <center>(MAX-MIN)/<b>numberOfNodes</b></center>
   * <br>
   * @param numberOfNodes Number of nodes in the network. 
   * @return The offset between two consecutive nodes, to get a network
   * with equidistance nodes. This static implementation always return null.
   * The expected value must be the same type as on the Id.setValue(Object) method.
   */
  public static Id divide(long numberOfNodes)
  { return null;}
  
  /**
   * This method returns the arithmetical result of this division:
   * <br>
   * <center>thisId/<b>divisor</b></center>
   * <br>
   * @param divisor The number of parts to divide this Id.
   * @return The result of division.
   */
  public abstract Id divideOn(long divisor);

}

