package planet.generic.commonapi.GML;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.Constructor;

import planet.commonapi.Id;
import planet.commonapi.GML.*;
import planet.commonapi.exception.InitializationException;

/**
 * Factory for the GML implementation classes.
 * @author <a href="mailto: marc.sanchez@estudiants.urv.es">Marc Sanchez</a>
 * @author <a href="mailto: jordi.pujol@estudiants.urv.es">Jordi Pujol</a>
 */
public class GMLFactoryImpl implements GMLFactory {
    
    /* *****  GML CLASS PROPERTIES  *************/
    /** Edge class implementation */
    public static final String DEFAULT_GML_EDGE_CLASS = "DEFAULT_GML_EDGE_CLASS";
    /** Default GMLConstraint implementation to use */
    public static final String DEFAULT_GML_CONSTRAINT_CLASS = "DEFAULT_GML_CONSTRAINT_CLASS";
    /** Default GMLConstraint implementation to use */
    public static final String DEFAULT_GML_GENERATOR_CLASS = "DEFAULT_GML_GENERATOR_CLASS";
  	
    
    /* ********* FACTORY PROPERTIES ***************************/
    /** Constructor for the GMLEdge, with the source, target and directed arguments.  */
    private transient Constructor edgeSrcTgtDirConstructor = null;
    /** Constructor for the GMLEdge, with the source, target, directed and fill arguments.  */
    private transient Constructor edgeSrcTgtDirFillConstructor = null;
    /** GMLEdge implementation class to use. */
    private Class edgeClass = null;
    /** GMLEGenerator implementation class to use. */
    private Class generatorClass = null;
    /** GMLConstraint implementation class to use. */
    private Class constraintClass = null;
    
  	/**
     * Loads all specified classes in the <b>propertiesFile</b>, related to the GML items. 
     * @param propertiesFile Path to the GML properties file.
     */
    public GMLFactoryImpl(String propertiesFile) throws InitializationException {
        if (propertiesFile==null) return;
		//Load properties
		FileInputStream fis = null;
		java.util.Properties props = new java.util.Properties();
		String propertiesFault = "";
		try {
			fis = new FileInputStream(propertiesFile);
			props.load(fis);
		} catch (FileNotFoundException e) {
			throw new InitializationException("Cannot read the file "+propertiesFile,e);
		} catch (IOException e) {
			throw new InitializationException("Cannot load the file "+propertiesFile,e);
		}
		
		//if all OK
		
		//read DEFAULT_GML_EDGE_CLASS
		try {
			edgeClass = Class.forName(props.getProperty(DEFAULT_GML_EDGE_CLASS));
			Class[] types  = {Id.class,Id.class,boolean.class};
			Class[] types2 = {Id.class,Id.class,boolean.class,String.class};
			edgeSrcTgtDirConstructor = edgeClass.getConstructor(types);
			edgeSrcTgtDirFillConstructor = edgeClass.getConstructor(types2);
		} catch (ClassNotFoundException e) {
			throw new InitializationException("Property "+DEFAULT_GML_EDGE_CLASS +" is not a valid class name.",e);
		} catch (Exception e) {
			throw new InitializationException("Cannot generate the GMLEdge initializers.",e);
		} 
		
		//read DEFAULT_GML_CONSTRAINT_CLASS
		try {
			constraintClass = Class.forName(props.getProperty(DEFAULT_GML_CONSTRAINT_CLASS));
		} catch (ClassNotFoundException e) {
			throw new InitializationException("Property "+DEFAULT_GML_CONSTRAINT_CLASS +" is not a valid class name.",e);
		}
		
		// read DEFAULT_GML_GENERATOR_CLASS
		try {
			generatorClass = Class.forName(props.getProperty(DEFAULT_GML_GENERATOR_CLASS));
		} catch (ClassNotFoundException e) {
			throw new InitializationException("Property "+DEFAULT_GML_GENERATOR_CLASS +" is not a valid class name.",e);
		}
		
    }
    
    /**
     * Build a GML edge with the specified values. The fill color is used the predefined one
     * (the appeared in GMLDefaultProperties.fill property).
     * @param source Source node Id.
     * @param target Destination node Id.
     * @param directed true when the edge is directed. false in other case.
     * @return An instance of GMLEdge.
     * @throws InitializationException if an error occurs during the initialization.
     * @see GMLDefaultProperties
     */
    public GMLEdge buildEdge(Id source, Id target, boolean directed) throws InitializationException
    {
        try {
	        Object params[] = {source,target,new Boolean(directed)};
	        return (GMLEdge)edgeSrcTgtDirConstructor.newInstance(params);
        } catch (Exception e)
        {
            throw new InitializationException("Cannot access to the constructor "+edgeClass+"(Id,Id,boolean)",e);
        }
    }
    
    /**
     * Build a GML edge with the specified values. The fill color is used the predefined one
     * (the appeared in GMLDefaultProperties.fill property).
     * @param source Source node Id.
     * @param target Destination node Id.
     * @param directed true when the edge is directed. false in other case.
     * @param fill Fill color in "#RRGGBB" format.
     * @return An instance of GMLEdge.
     * @throws InitializationException if an error occurs during the initialization.
     */
    public GMLEdge buildEdge(Id source, Id target, boolean directed, String fill) throws InitializationException
    {
        try {
	        Object params[] = {source,target,new Boolean(directed),fill};
	        return (GMLEdge)edgeSrcTgtDirFillConstructor.newInstance(params);
        } catch (Exception e)
        {
            throw new InitializationException("Cannot access to the constructor "+edgeClass+"(Id,Id,boolean,String)",e);
        }
    }

    
    /**
     * Builds the GMLGenerator specified in the GML properties file.
     * @return The GMLGenerator.
     * @throws InitializationException if an error occurs during the initialization.
     */
    public GMLGenerator buildGMLGenerator() throws InitializationException 
    {
        try {
            return (GMLGenerator)generatorClass.newInstance();
        } catch (Exception e) {
            throw new InitializationException("Cannot access to the empty constructor "+generatorClass+"()",e);
        }
    }
    
    /**
     * Build the GMLConstraint specified in the GML properties file.
     * @return The GMLConstraint.
     * @throws InitializationException if an error occurs during the initialization.
     */
    public GMLConstraint buildDefaultGMLConstraint() throws InitializationException 
    {
        try {
            return (GMLConstraint)constraintClass.newInstance();
        } catch (Exception e) {
            throw new InitializationException("Cannot access to the empty constructor "+constraintClass+"()",e);
        }        
    }
}
