package planet.generic.commonapi.factory;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.Constructor;


import planet.util.Properties;
import planet.commonapi.Id;
import planet.commonapi.NodeHandle;
import planet.commonapi.exception.InitializationException;
import planet.commonapi.factory.*;

/**
 * This is a specific implementation of NodeHandleFactory that permits
 * build any class of NodeHandle with any type of Id. It is possible by
 * identifies this parameter in properties file:
 * <ol>
 * <li><b>DEFAULT_NODEHANDLECLASS</b>: Class which identifies the NodeHandle to build.</li> 
 * </ol>
 * It can be modified too by invoking method to this class. In any case, 
 * the properties file only can be modified by programmer.
 * <br><br>
 * IMPORTANT: The concrete implementation of NodeHandle must have one constructor:
 * <ol>
 * <li>The constructor with the related Node Id in its first parameter, and a 
 * boolean to show if the Node is alive, at the sencod argument.</li>
 * <ol>
 * @author Jordi Pujol
 * Date: 13/05/04
 */
public class NodeHandleFactoryImpl implements NodeHandleFactory {
	/**
	 * Actual instance of Class for new Nodes.
	 */
	private Class nodeHandleClass = null;
	/**
	 * Constructor of actual nodeClass with only one parameter: an Id object.
	 */
	private transient Constructor nodeHandleClassConstructor = null;
	
	
	/**
	 * This constructor reads the properties' file and initializes
	 * the Class for new NodeHandles. If occur any problem
	 * with this initialization, throws a InitializationException 
	 * exception with short description for the problem.
	 * @param propertiesFile File of properties with the <b>DEFAULT_NODEHANDLECLASS</b>
	 * property.
	 * @throws InitializationException If occur any problem during
	 * initialization steps.
	 */
	public NodeHandleFactoryImpl(String propertiesFile) throws InitializationException {
		//Load properties
		FileInputStream fis = null;
		java.util.Properties nodeProp = new java.util.Properties();
		try {
			fis = new FileInputStream(propertiesFile);
			nodeProp.load(fis);
		} catch (FileNotFoundException e) {
			throw new InitializationException("Cannot read the file "+propertiesFile,e);
		} catch (IOException e) {
			throw new InitializationException("Cannot load the file "+propertiesFile,e);
		}
		
		//if all OK
		
		//read Properties.DEFAULT_NODEHANDLECLASS
		try {
			nodeHandleClass = Class.forName(nodeProp.getProperty(Properties.DEFAULT_NODEHANDLECLASS));
			Class[] types = {planet.commonapi.Id.class,boolean.class};
			nodeHandleClassConstructor = nodeHandleClass.getConstructor(types);
		} catch (ClassNotFoundException e) {
			throw new InitializationException("Property ["+Properties.DEFAULT_NODEHANDLECLASS +"] is not a valid class name.",e);
		} catch (Exception e) {
			throw new InitializationException("Cannot gets the constructor for class ["+nodeHandleClass
					+"] of property ["+Properties.DEFAULT_NODEHANDLECLASS +"]",e);
		} 
	}
	
	/**
	 * Initializes a new instance of NodeHandleFactory with the 
	 * <b>nodeHandleClass</b> specified by parameter.
	 * If exists any problem with this class an exception is thrown
	 * with a short description for the problem.
	 * @param nodeHandleClass Class for new NodeHandles.
	 * @throws InitializationException If exists any problem with
	 * initialization of class.
	 * @see planet.commonapi.Id Id
	 */
	public NodeHandleFactoryImpl(Class nodeHandleClass) throws InitializationException {
		try {
			this.nodeHandleClass = nodeHandleClass;
			Class[] types = {planet.commonapi.Id.class,boolean.class};
			nodeHandleClassConstructor = nodeHandleClass.getConstructor(types);
		} catch (Exception e) {
			throw new InitializationException("Cannot gets the constructor for class ["+nodeHandleClass
					+"] of property ["+Properties.DEFAULT_NODEHANDLECLASS +"]",e);
		}

	}

	/**
	 * Builds a NodeHandle with the actual implementation class of NodeHandle and
	 * specifieds <b>id</b> and <b>alive</b> flag.
	 * @param id Id of the related Node, to assign to the new instance of NodeNandle.
	 * @param alive true if the related Node is alive. false in other case.
	 * @return An instance of actual implementation class of NodeHandle.
	 * @see planet.commonapi.factory.NodeHandleFactory#buildNodeHandle(planet.commonapi.Id, boolean)
	 * @see planet.commonapi.Id Id
	 */
	public NodeHandle buildNodeHandle(Id id, boolean alive) throws InitializationException {
		NodeHandle nodeHandle = null;
		Object params[] = {id,Boolean.valueOf(alive)};
		try {
			nodeHandle = (NodeHandle) nodeHandleClassConstructor.newInstance(params);
		} catch (IllegalArgumentException e) {
			throw new InitializationException("Bad arguments at Constructor "+nodeHandleClass.getName()+"(Id,boolean)",e);
		} catch (Exception e) {
			throw new InitializationException("Instantiation problems at Constructor "+nodeHandleClass.getName()+"(Id,boolean)",e);
		}
		return nodeHandle;
	}
	
	/**
	 * Gets the actual class of implementation of NodeHandle returned by this factory.
	 * @see planet.commonapi.factory.NodeHandleFactory#getNodeHandleClass()
	 * @return The actual class of implementation of NodeHandle.
	 */
	public Class getNodeHandleClass() {
		return nodeHandleClass;
	}
	
	/**
	 * Sets the class of an implementation of NodeHandle to return by this
	 * factory in the future.
	 * @see planet.commonapi.factory.NodeHandleFactory#setNodeHandleClass(java.lang.Class)
	 * @param nodeHandleClass Class of an implementation of NodeHandle to return by
	 * this factory.
	 */
	public void setNodeHandleClass(Class nodeHandleClass) throws InitializationException {
		this.nodeHandleClass = nodeHandleClass;
		try {
			Class types[] = {Id.class,boolean.class};
			nodeHandleClassConstructor = nodeHandleClass.getConstructor(types);
		} catch (SecurityException e) {
			throw new InitializationException("Cannot access to the Constructor "+nodeHandleClass.getName()+"(Id,boolean)",e);
		} catch (NoSuchMethodException e) {
			throw new InitializationException("Not exists Constructor "+nodeHandleClass.getName()+"(Id,boolean)",e);
		}
	}
	
	/**
	 * Makes nothing special, only invokes to <b>stream.defaultWriteObject()</b>.
	 * @param stream Stream to save the actual instance.
	 * @throws IOException if any error has ocurred.
	 */
	private void writeObject(java.io.ObjectOutputStream stream)
    throws IOException {
		stream.defaultWriteObject();
	}

	/**
	 * Makes a first <b>stream.defaultReadObject()</b> invocation to uses
	 * the default read method. And then, builds the constructor for
	 * the loaded NodeHandle class.
	 * @param stream Stream to read the actual instance.
	 * @throws IOException if occur any error during 
	 * default read object, or if there are any error during
	 * building of NodeHandle constructor.  
	 * @throws ClassNotFoundException
	 */
	private void readObject(java.io.ObjectInputStream stream)
    throws IOException, ClassNotFoundException {
		stream.defaultReadObject();
		try {
			Class types[] = {Id.class, boolean.class};
			nodeHandleClassConstructor = this.nodeHandleClass.getConstructor(types);
		} catch (SecurityException e) {
			throw new IOException("Cannot access to the Constructor "+nodeHandleClass+"(Id,boolean)");
		} catch (NoSuchMethodException e) {
			throw new IOException("Not exists Constructor "+nodeHandleClass+"(Id,boolean)");
		}
	}
}
