package planet.symphony;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import planet.commonapi.exception.InitializationException;
import planet.util.OverlayProperties;

/**
 * This class includes the initialization and the values for all
 * configuration properties of the Chord overlay.
 * @author <a href="mailto: jordi.pujol@estudiants.urv.es">Jordi Pujol</a>
 * @author <a href="mailto: marc.sanchez@estudiants.urv.es">Marc Sanchez</a>
 * Date: 05/07/2004
 */
public class SymphonyProperties implements OverlayProperties {

	/* *********************** SYMPHONY PROPERTIES ***********************/
	/* Theese must to appear on the properties file */
	
	/**
	 * Symphony property: Default key for maximum number of long distance 
	 * connections.
	 */
	public static final String DEFAULT_K = "DEFAULT_K";
	/**
	 * Symphony property: Default key for maximum number of members
	 * in successor list.
	 */
	public static final String DEFAULT_F = "DEFAULT_F";
	/**
	 * Symphony property: Default key for maximum number of retries
	 * to obtain a connection to the same long distance node.
	 */
	public static final String DEFAULT_MAX_RETRIES_NEW_LONG_DISTANCE = "DEFAULT_MAX_RETRIES_NEW_LONG_DISTANCE";
	/**
	 * Symphony property: Default key for maximum number of retries
	 * to enter to the network by the same bootstrap.
	 */
	public static final String DEFAULT_MAX_JOIN_RETRIES = "DEFAULT_MAX_JOIN_RETRIES";
	/**
	 * Symphony property: Default key for number of stabilize steps.
	 */
	public static final String DEFAULT_STABILIZE_STEPS = "DEFAULT_STABILIZE_STEPS";
		
	/* *************** SYMPHONY ATTRIBUTES   ****************/
	/**
	 * Maximum number of long distance connections.
	 */
	public int K;
	/**
	 * Maximum number of members in successor list.
	 */
	public int F;
	/**
	 * Maximum number of retries
	 * to obtain a connection to the same long distance node.
	 */
	public int maxRetriesNewLongDistance;
	/**
	 * Maximum number of retries
	 * to enter to the network by the same bootstrap.
	 */
	public int maxJoinRetries;
	/**
	 * Number of steps for statibilization.
	 */
	public int stabilizeSteps;


	/**
	 * Initialize all configuration properties of the Symphony overlay.
	 * @see planet.util.OverlayProperties#init(java.lang.String)
	 * @param propertiesFile
	 * @throws InitializationException
	 */
	public void init(String propertiesFile) throws InitializationException {
		//Load properties
		FileInputStream fis = null;
		java.util.Properties prop = new java.util.Properties();
		try {
			fis = new FileInputStream(propertiesFile);
			prop.load(fis);
		} catch (FileNotFoundException e) {
			throw new InitializationException("Cannot read the file "+propertiesFile,e);
		} catch (IOException e) {
			throw new InitializationException("Cannot load the file "+propertiesFile,e);
		}
		
		//if all OK
	

		//read the K
		K = Integer.parseInt(prop.getProperty(DEFAULT_K));
		if (K < 0) 
			throw new InitializationException("Property ["+DEFAULT_K +"] is not a valid number.");
		
		//read the F
		F = Integer.parseInt(prop.getProperty(DEFAULT_F));
		/*
		 * IMPORTANT: DEFAULT_F>= 2
		 * To calculate correctly the number of members in the 
		 * network each node requires the predecessor and
		 * the predecessor of its predecessor. For this,
		 * the value of this property must be greater than 
		 * or equals to two (2).
		 */
		if (F < 2) 
			throw new InitializationException("Property ["+DEFAULT_F +"] is not a valid number.");

		//read the maxRetriesNewLongDistance
		maxRetriesNewLongDistance = Integer.parseInt(prop.getProperty(DEFAULT_MAX_RETRIES_NEW_LONG_DISTANCE));
		if (maxRetriesNewLongDistance < 0) 
			throw new InitializationException("Property ["+DEFAULT_MAX_RETRIES_NEW_LONG_DISTANCE +"] is not a valid number.");
		
		//read the maxJoinRetries
		this.maxJoinRetries = Integer.parseInt(prop.getProperty(DEFAULT_MAX_JOIN_RETRIES));
		if (maxJoinRetries < 0) 
			throw new InitializationException("Property ["+DEFAULT_MAX_JOIN_RETRIES +"] is not a valid number.");

		//read the defaultStabilizeSteps
		stabilizeSteps = Integer.parseInt(prop.getProperty(DEFAULT_STABILIZE_STEPS));
		if (stabilizeSteps < 0) 
			throw new InitializationException("Property ["+DEFAULT_STABILIZE_STEPS +"] is not a valid number of stabilize steps.");
		
	}
	
	
	/**
	 * Returns a String representation of the constant specific values
	 * of type the RouteMessage. Its use is only for human readable 
	 * logs. Based on SymphonyNode implementation.
	 * @param type Value to get its String representation.
	 * @return The String representation of the type.
	 */
	public String typeToString(int type) {
		return SymphonyNode.TYPES[type];
	}
	/**
	 * Returns a string representation of each of event mode and 
	 * RouteMessage mode.
	 * @param mode Mode of the RouteMessage to get its 
	 * String representation.
	 * @return String representation of the mode of RouteMessage.
	 */
	public String modeToString(int mode) {
		return SymphonyNode.MODES[mode];
	}
	
	/**
	 * Returns RouteMessage type for Application level.
	 * @return RouteMessage type for Application level.
	 */
	public int getTypeForApplicationMessage()
	{
	    return SymphonyNode.DATA;
	}
}
